package dev.pfaff.altargui.glfw;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.GuiGlobalsInit;
import dev.pfaff.altargui.Shaders;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.glfw.ring.LongRingBuffer;
import dev.pfaff.altargui.glfw.ring.Stats;
import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMod;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.altargui.text.FontManager;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.altargui.util.ThreadFuture;
import dev.pfaff.glass.OS;
import dev.pfaff.glass.apple.gcd.GrandCentralDispatch;
import dev.pfaff.glass.glfw.*;
import dev.pfaff.glass.opengl.GLDebugUtil;
import dev.pfaff.glass.opengl.enums.GLEnableCap;
import dev.pfaff.log4truth.LogFilter;
import dev.pfaff.log4truth.LogSink;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.unfettered.Unfettered;
import org.jspecify.annotations.Nullable;

import java.lang.ref.Reference;
import java.util.BitSet;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Supplier;

import static dev.pfaff.altargui.glfw.Config.*;
import static dev.pfaff.altargui.glfw.GLGlobals.GL;
import static dev.pfaff.altargui.glfw.GLGlobals.GLFW;
import static dev.pfaff.log4truth.StandardTags.*;
import static java.lang.invoke.MethodType.methodType;
import static java.util.FormatProcessor.FMT;

// TODO: refactor Application things be `static final`, separate from window, provide facilities for (multi-)window
//  creation
public final class Application {
	static {
		Unfettered.forceInit();

//		Module javaBaseModule = String.class.getModule();
//		Module ourModule = Main.class.getModule();
//		Unfettered.export(javaBaseModule, "jdk.internal.ref", false, ourModule);
	}

	private final ConcurrentLinkedQueue<Runnable> tasks = new ConcurrentLinkedQueue<>();

	void submitToMainThreadOrRun(Runnable task) {
		if (Thread.currentThread() == this.mainThread) {
			task.run();
		} else {
			this.tasks.add(task);
		}
	}

	private static final LongRingBuffer timingTasks = new LongRingBuffer(1024 * 10);
	private static final LongRingBuffer timingRender = new LongRingBuffer(1024 * 10);
	private static final LongRingBuffer timingFlushFramebuffer = new LongRingBuffer(1024 * 10);
	private static final LongRingBuffer timingDisplayUpdate = new LongRingBuffer(1024 * 10);
	private static final LongRingBuffer timingFrame = new LongRingBuffer(1024 * 10);
	private static final Stats stats = new Stats()
			.timing("Tasks", timingTasks)
			.timing("Render", timingRender)
			.timing("Flush framebuffer", timingFlushFramebuffer)
			.timing("Display update", timingDisplayUpdate)
			.timing("Frame", timingFrame);

//	private static final int GL_CLEAR_COLOR = 0xfefefeff;
//	private static GLClear GL_CLEAR;

//	private static void applyStaticBuffer() {
//		GL_CLEAR = new GLClear(hglBuffer.bb().slice(0, GLClear.LAYOUT.length()));
//		GL_CLEAR.init();
//		GL_CLEAR.color(GL_CLEAR_COLOR);
//	}

	private final Thread mainThread = Thread.currentThread();
	final GLFWwindow window;
	@SuppressWarnings("FieldCanBeLocal")
	private final WindowCallbackManager callbackManager;
	private final GLFWUIUtil glfwUIUtil;
	private final DelayedInitGuiContext gui;

	private boolean windowIconified = false;

	private static boolean logFilterAnd(LogFilter first, LogFilter second, String source, List<String> tags, Thread thread) {
		return first.test(source, tags, thread) && second.test(source, tags, thread);
	}
	private static boolean logFilterOr(LogFilter first, LogFilter second, String source, List<String> tags, Thread thread) {
		return first.test(source, tags, thread) || second.test(source, tags, thread);
	}
	private static boolean logFilterNot(LogFilter filter, String source, List<String> tags, Thread thread) {
		return !filter.test(source, tags, thread);
	}

	private Application(Supplier<Widget> windowContentSupplier) {
		GuiGlobalsInit.setHost(new HostImpl(this));
		GLThread.setSubmitToGLThread(this::submitToMainThreadOrRun);

		GLGlobals.forceInit();

		Thread.startVirtualThread(FontManager::reload);

		var filter = System.getenv("LOG_FILTER");
		if (filter != null) {
			LogFilter cases = null;
			var i = 0;
			do { // each case
				LogFilter conditions = null;
				var caseEnd = filter.indexOf(';', i);
				if (caseEnd == -1) caseEnd = filter.length();
				var j = i;
				while (j < caseEnd) { // each condition (AND)
					var condEnd = filter.indexOf(',', j, caseEnd);
					if (condEnd == -1) condEnd = caseEnd;

					var fieldNameEnd = filter.indexOf('=', j, condEnd);
					if (fieldNameEnd == -1) {
						throw new IllegalArgumentException(STR."Invalid log filter: expected a '=' somewhere between in the range [\{j}, \{condEnd})");
					}

					enum Field {
						Tag,
						Source,
					}
					var field = filter.regionMatches(j, "tag", 0, fieldNameEnd - j)
							? Field.Tag
							: filter.regionMatches(j, "source", 0, fieldNameEnd - j)
							? Field.Source
							: null;

					var condition = switch (field) {
						case null -> {
							var fieldName = filter.substring(j, fieldNameEnd);
							throw new IllegalArgumentException(STR."Invalid log filter: unrecognized field name at [\{j}, \{fieldNameEnd}): \"\{fieldName}\"");
						}
						case Tag -> {
							var excludeTag = filter.substring(fieldNameEnd+1, condEnd);
							yield (LogFilter) (_, tags, _) -> tags.contains(excludeTag);
						}
						case Source -> {
							var excludeSource = filter.substring(fieldNameEnd+1, condEnd);
							yield (LogFilter) (source, _, _) -> source.equals(excludeSource);
						}
					};
					if (conditions == null) {
						conditions = condition;
					} else {
						conditions = conditions.and(condition);
					}

					j = condEnd + 1;
				}

				if (conditions == null) {
					throw new IllegalArgumentException(STR."Invalid log filter: no conditions in case at [\{i}, \{caseEnd})");
				}
				if (cases == null) {
					cases = conditions.negate();
				} else {
					cases = cases.and(conditions.negate());
				}

				i = caseEnd + 1;
			} while (i < filter.length());

			LogSink.addFilterToGlobal(cases);
		} else {
			LogSink.addFilterToGlobal((_, tags, _) -> !tags.contains("DEBUG"));
		}

		GLFW.setErrorCallback((code, description) -> {
			Logger.log(() -> STR."GLFW error \{code}: \{description}", List.of(ERROR));
		});

////		var dm = Display.getDisplayMode();
////		var bpp = dm.getBitsPerPixel();
//		var dm = findDisplayMode(_ -> true);
//		if (dm == null) dm = Display.getDesktopDisplayMode();
//		try {
////			Display.setDisplayModeAndFullscreen(dm);
//			Display.setDisplayMode(dm);
//		} catch (LWJGLException e) {
//			throw new RuntimeException(e);
//		}
//		var w = 350;
//		var h = 200;
//		var minX = (dm.getWidth() - w) / 2;
//		var minY = (dm.getHeight() - h) / 2;
//		Display.setLocation(minX, minY);
//		Display.setTitle("OpenGL Stuff");
//		Display.setResizable(true);
//		try {
////			Display.create(new PixelFormat().withSamples(4));
//			Display.create(new PixelFormat(), new ContextAttribs(4, 1, 1, 1 | 2));
//		} catch (LWJGLException e) {
//			throw new RuntimeException(e);
//		}
		GLFWwindow.setBooleanHint(GLFW, WindowHint.SCALE_FRAMEBUFFER, SCALE_FRAMEBUFFER);
		GLFWwindow.setBooleanHint(GLFW, WindowHint.SCALE_TO_MONITOR, SCALE_TO_MONITOR);
		GLFWwindow.setBooleanHint(GLFW, WindowHint.DOUBLEBUFFER, DOUBLE_BUFFER);
		GLFWwindow.setIntegerHint(GLFW, WindowHint.SAMPLES, 16);

		// try full OpenGL first, then fall-back to OpenGL ES
		GLFWwindow.setIntegerHint(GLFW, WindowHint.CONTEXT_VERSION_MAJOR, 4);
		GLFWwindow.setIntegerHint(GLFW, WindowHint.CONTEXT_VERSION_MINOR, 1);
		GLFWwindow.setIntegerHint(GLFW, WindowHint.CLIENT_API, ClientAPI.OPENGL.value());
		GLFWwindow.setIntegerHint(GLFW, WindowHint.OPENGL_PROFILE, OpenGLProfile.CORE_PROFILE.value());
		GLFWwindow.setBooleanHint(GLFW, WindowHint.VISIBLE, false);

		GLFWwindow window1;
		try {
			window1 = createWindow();
		} catch (RuntimeException e) {
			GLFWwindow.setIntegerHint(GLFW, WindowHint.CONTEXT_VERSION_MAJOR, 3);
			GLFWwindow.setIntegerHint(GLFW, WindowHint.CONTEXT_VERSION_MINOR, 1);
			GLFWwindow.setIntegerHint(GLFW, WindowHint.CLIENT_API, ClientAPI.OPENGL_ES.value());
			GLFWwindow.setIntegerHint(GLFW, WindowHint.OPENGL_PROFILE, OpenGLProfile.ANY_PROFILE.value());
			window1 = createWindow();
		}
		window = window1;
		window.makeContextCurrent();
		GLFWwindow.setSwapInterval(GLFW, V_SYNC ? 1 : 0);

		// ensure GuiGlobals is initialized on the main thread
		// must be done after the OpenGL context has been created
		GuiGlobals.forceInit();
		// delayed until after GuiGlobals is initialized
		var shaderLoadThread = ThreadFuture.spawnVirtualTask(() -> {
			Shaders.reload();
			DelayedInitGuiContext.shadersLoaded();
		});

		GL.glEnable(GLEnableCap.MULTISAMPLE);

		GLDebugUtil.logGLInfo(GL);

//		GL.glMatrixMode(GLMatrixMode.PROJECTION);
//		GL.glLoadIdentity();
//		gluOrtho2D(0, dm.getWidth(), dm.getHeight(), 0);
//		GL.glMatrixMode(GLMatrixMode.MODELVIEW);
//		GL.glLoadIdentity();
//		var dm_ = dm;
//		Logger.log(() -> "display mode: " + dm_);
//		Logger.log(() -> "display: width=" + Display.getWidth() + ", height=" + Display.getHeight());
//		GL.glViewport(0, 0, dm.getWidth(), dm.getHeight());
//		var white = 254 / 255f;
//		GL.glClearColor(white, white, white, 1f);

//        try {
//            Display.setVSyncEnabled(true);
//        } catch (LWJGLException e) {
//            throw new RuntimeException(e);
//        }

//		applyStaticBuffer();

		gui = DelayedInitGuiContext.create(windowContentSupplier.get());

		callbackManager = new WindowCallbackManager(window);
		callbackManager.setKeyCallback((_, keyCode, scanCode, action, modifierBits) -> {
			// keyCode is -1 when a modifier key is pressed/released
			// TODO: do we want to handle these somehow?
			if (keyCode == -1) return;
			var key = InputKey.fromCode(keyCode);
			if (key == null) {
				Logger.log(() -> "Unrecognized key code " + keyCode, List.of(WARN));
				return;
			}
			var modifiers = InputMods.fromGLFW(modifierBits);
			if (action == KeyAction.RELEASE.value()) {
				gui.keyReleased(key, scanCode, modifiers);
			} else {
				if (action == KeyAction.REPEAT.value()) {
					modifiers = modifiers.set(InputMod.REPEAT);
				}
				gui.keyPressed(key, scanCode, modifiers);
			}
		});
		callbackManager.setCharCallback((_, codePoint) -> gui.charTyped(codePoint));
		var windowState = new Object() {
			private GLFWwindow.ContentScale contentScale = new GLFWwindow.ContentScale(0f, 0f);
		};
		var mouseState = new Object() {
			private double x, y;
			private final BitSet activeMouseButtons = new BitSet(8);
		};
		callbackManager.setCursorPosCallback((_, xIn, yIn) -> {
//			var x = xIn * windowState.contentScale.x();
//			var y = yIn * windowState.contentScale.y();
			var glfwAltarGUIUtil = glfwAltarGUIUtil();
			if (glfwAltarGUIUtil != null) {
				var scaleFactor = glfwAltarGUIUtil().screenCoordinatesToVirtualPixelsScaleFactor();
				var x = xIn * scaleFactor;
				var y = yIn * scaleFactor;

				var deltaX = x - mouseState.x;
				var deltaY = y - mouseState.y;
				mouseState.x = x;
				mouseState.y = y;

				gui.mouseMoved(x, y);
				mouseState.activeMouseButtons.stream().forEach(button -> {
					gui.mouseDragged(x, y, button, deltaX, deltaY);
				});
			}
		});
		callbackManager.setScrollCallback((_, xOffset, yOffset) -> {
			// TODO: use virtual scale factor
			gui.mouseScrolled(mouseState.x, mouseState.y, yOffset * windowState.contentScale.y());
		});
		callbackManager.setMouseButtonCallback((_, button, action, modifiers) -> {
			if (action == MouseButtonAction.RELEASE.value()) {
				gui.mouseReleased(mouseState.x, mouseState.y, button);
				mouseState.activeMouseButtons.clear(button);
			} else {
				gui.mousePressed(mouseState.x, mouseState.y, button);
				mouseState.activeMouseButtons.set(button);
			}
		});
		callbackManager.setPathDropCallback((_, paths) -> {
			// TODO: implement
		});
		callbackManager.setWindowSizeCallback((_, width, height) -> {
			Logger.log(() -> STR."Received new window size: width=\{width}, height=\{height}", List.of(DEBUG));
			if (width == 0 || height == 0) {
				Logger.log(() -> "Ignoring width or height of 0", List.of(DEBUG));
				return;
			}
			if (Thread.currentThread() == this.mainThread) {
				var glfwAltarGUIUtil = glfwAltarGUIUtil();
				if (glfwAltarGUIUtil != null) {
					glfwAltarGUIUtil.windowSizeChanged(width, height);
				}
			} else {
				Logger.log(() -> "Received new window size on non-main thread. The implications of this are " +
								 "unknown. The window may not render until the user ceases resizing.",
						List.of(WARN));
			}
		});
		callbackManager.setFramebufferSizeCallback((_, width, height) -> {
			Logger.log(() -> STR."Received new framebuffer size: width=\{width}, height=\{height}", List.of(DEBUG));
			if (width == 0 || height == 0) {
				Logger.log(() -> "Ignoring width or height of 0", List.of(DEBUG));
				return;
			}
			if (Thread.currentThread() == this.mainThread) {
				var glfwAltarGUIUtil = glfwAltarGUIUtil();
				if (glfwAltarGUIUtil != null) {
					glfwAltarGUIUtil.framebufferSizeChanged(width, height);
				}
				GL.glViewport(0, 0, width, height);
				render(window);
			} else {
				Logger.log(() -> "Received new framebuffer size on non-main thread. The implications of this are " +
								 "unknown. The window may not render until the user ceases resizing.",
						List.of(WARN));
			}
//			if (width == 1680 && height == 1050) {
//				Logger.log(() -> "Setting window size");
//				window.setWindowSize(2880, 1800);
//			}
		});
		callbackManager.setWindowContentScaleCallback((_, x, y) -> {
			Logger.log(() -> STR."Received new window content scale: x=\{x}, y=\{y}", List.of(DEBUG));
			var glfwAltarGUIUtil = glfwAltarGUIUtil();
			if (glfwAltarGUIUtil != null) {
				glfwAltarGUIUtil.contentScaleChanged(x, y);
			}
			windowState.contentScale = new GLFWwindow.ContentScale(x, y);
		});
		callbackManager.setWindowMaximizeCallback((_, maximized) -> {
			if (maximized) {
				Logger.log(() -> "Maximized");
			} else {
				Logger.log(() -> "Un-maximized");
			}
		});
		callbackManager.setWindowIconifyCallback((_, iconified) -> {
			if (iconified) {
				Logger.log(() -> "Iconified");
			} else {
				Logger.log(() -> "Un-iconified");
			}
			this.windowIconified = iconified;
		});
		callbackManager.setWindowCloseCallback(_ -> {
			// TODO: gui event
		});

		// This call is necessary for the following window.get* calls to return correct
		// information. For whatever reason, despite changing what those calls return, no events
		// will be fired as you might expect.
		GLFW.pollEvents();
		glfwUIUtil = new GLFWUIUtil(window, gui);
		windowState.contentScale = window.getWindowContentScale();
		Logger.log(() -> STR."""
			Initial window properties:
				Content scale: \{windowState.contentScale}
				Window size: \{window.getWindowSize()}
				Framebuffer size: \{window.getFramebufferSize()}""");
		this.windowIconified = window.getBooleanWindowAttrib(WindowAttrib.ICONIFIED);

		while (shaderLoadThread.isAlive()) {
			runTasks();
		}
		shaderLoadThread.join();

		// testing shows that it is optimal to delay showing the window until after the initial call to GLFW.pollEvents
		window.setVisibility(true);
	}

	private static GLFWwindow createWindow() {
		return GLFWwindow.create(GLFW, 800, 600, "Altar GUI Demo", null, null);
	}

	/**
	 * Accessor provides static analysis with the information that the field could be {@code null}.
	 */
	@SuppressWarnings("DataFlowIssue")
	@Nullable
	private GLFWUIUtil glfwAltarGUIUtil() {
		return glfwUIUtil;
	}

	/**
	 * @param windowContentSupplier a supplier of the window's content {@link Widget}. This is supplied so that
	 *                              necessary global initialization can be performed before loading widget classes.
	 */
	public static void run(Supplier<Widget> windowContentSupplier) {
		Thread.currentThread().setName(Thread.currentThread().getName() + "-boot");
		Logger.log(() -> "Running application", List.of(INFO));

		if (OS.HOST instanceof OS.macOS) {
			GrandCentralDispatch.getMainQueue().dispatchSync(() -> runOnThisThread(windowContentSupplier));
		} else {
			runOnThisThread(windowContentSupplier);
		}

		System.out.println();
		System.out.println(stats.makeReport());
	}

	private static void runOnThisThread(Supplier<Widget> windowContentSupplier) {
		Thread.currentThread().setName(Thread.currentThread().getName() + "-main");
		Logger.log(() -> "The app will run on this thread", List.of(INFO));

		var main = new Application(windowContentSupplier);
		main.loop();
		long start = System.nanoTime();
		main.window.destroy();
		GLFW.close();
		Reference.reachabilityFence(main.callbackManager);
		long elapsed = System.nanoTime() - start;
		System.out.println(FMT."Destroyed window and terminated GLFW in %.2f\{elapsed / 1_000_000.0} ms");
	}

//	/**
//	 * Finds the first display mode that passes the predicate, or null.
//	 */
//	@Nullable
//	private static DisplayMode findDisplayMode(Predicate<DisplayMode> predicate) {
//		DisplayMode[] modes;
//		try {
//			modes = Display.getAvailableDisplayModes();
//		} catch (LWJGLException e) {
//			Logger.log(() -> "Unable to determine available display modes", e, List.of(ERROR));
//			return null;
//		}
//		Logger.log(() -> "Available display modes: " + Arrays.toString(modes));
//		Arrays.sort(modes, Comparator.comparingInt(DisplayMode::getBitsPerPixel).reversed()
//				.thenComparing(Comparator.comparingInt(DisplayMode::getFrequency).reversed())
//				.thenComparing(Comparator.comparingInt(DisplayMode::getWidth).reversed())
//				.thenComparing(Comparator.comparingInt(DisplayMode::getHeight).reversed()));
//		for (var mode : modes) {
//			if (predicate.test(mode)) return mode;
//		}
//		return null;
//	}

	private static final long FRAME_STAT_INTERVAL = 2_500_000_000L;

	private static long nextFrameStatTarget(long now) {
		return now + FRAME_STAT_INTERVAL;
	}

	private void runTasks() {
		long start = System.nanoTime(), now;
		Runnable task;
		while ((task = tasks.poll()) != null) task.run();
		now = System.nanoTime();
		timingTasks.add(now - start);
	}

	private void loop() {
		int frames = 0;
		long target = nextFrameStatTarget(System.nanoTime());

		while (!window.shouldClose()) {
			long frameStart = System.nanoTime();
			long now, start;

			runTasks();

			// skip render when iconified (minimized)
			if (!windowIconified) {
				render(window);
				++frames;
			}

			start = System.nanoTime();
			GLFW.pollEvents();
			timingDisplayUpdate.add(System.nanoTime() - start);

//			while (Keyboard.next()) {
//				var key = InputKey.fromCode(Keyboard.getEventKey());
//
//			}

			now = System.nanoTime();
			if (now > target) {
				var now_ = now;
				var frames_ = frames;
				var target_ = target;
				Logger.log(() -> STR."\{frames_ / (double) (FRAME_STAT_INTERVAL + (now_ - target_)) *
										1_000_000_000} fps (\{frames_} frames in the last 2.5 seconds)");
				frames = 0;
				target = nextFrameStatTarget(now);
			}
			timingFrame.add(System.nanoTime() - frameStart);
		}
	}

	private void render(GLFWwindow window) {
		long start = System.nanoTime();
		gui.render();
		long now = System.nanoTime();
		timingRender.add(now - start);

		start = now;

		if (DOUBLE_BUFFER) {
			window.swapBuffers();
		} else {
			GL.glFlush();
//			GL.glFinish();
		}
		now = System.nanoTime();
		timingFlushFramebuffer.add(now - start);
	}
}
