package dev.pfaff.altargui.glfw;

import dev.pfaff.altargui.GuiContext;
import dev.pfaff.altargui.Shaders;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.altargui.text.FontManager;
import dev.pfaff.log4truth.Logger;
import org.jspecify.annotations.Nullable;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MutableCallSite;

import static java.lang.invoke.MethodHandles.constant;

public final class DelayedInitGuiContext {
	public static final MethodHandle MH_GuiContext_null = constant(GuiContext.class, null);

	private final MutableCallSite guiContext = new MutableCallSite(MH_GuiContext_null);
	private final Widget widget;

	/**
	 * Creates a new UI instance, bound to the specified widget.
	 *
	 * @return the new UI instance.
	 */
	// Using a factory in case I want to change the implementation in the future.
	public static DelayedInitGuiContext create(Widget widget) {
		return new DelayedInitGuiContext(widget);
	}

	private DelayedInitGuiContext(Widget widget) {
		this.widget = widget;
	}

	private static boolean canInitialize() {
		return Shaders.isInitialized() && FontManager.isInitialized();
	}

	public int renderBufferByteCount() {
		var context = getContext();
		if (context == null) return 0;
		return context.renderBufferByteCount();
	}

	private int width;
	private int height;
	private float contentScale;

	/**
	 * Must be called on the GL thread.
	 */
	public void onWindowResize(int width, int height, float contentScale) {
		this.width = width;
		this.height = height;
		this.contentScale = contentScale;

		var xScaleFactor = (float) ((2.0 / width) * contentScale);
		var yScaleFactor = (float) ((-2.0 / height) * contentScale);
		Logger.log(() -> STR."Shader scale factor: \{xScaleFactor}, \{yScaleFactor}");
		Shaders.updateUniforms(xScaleFactor, yScaleFactor);

		var context = getContext();
		if (context != null) {
			context.setWindowProps(contentScale, width, height);
		}
	}

	public static void shadersLoaded() {
//		GLThread.submitToGLThread(() -> onWindowResize(width, height, contentScale));
	}

	@Nullable
	private GuiContext getContext() {
		if (!canInitialize()) return null;
		try {
			return (GuiContext) guiContext.getTarget().invokeExact();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	@Nullable
	private GuiContext getOrInitializeContext() {
		if (!canInitialize()) return null;
		var context = getContext();
		if (context == null) {
			return initializeContext();
		}
		return context;
	}

	private GuiContext initializeContext() {
		var context = new GuiContext(widget);
		guiContext.setTarget(constant(GuiContext.class, context));
		onWindowResize(width, height, contentScale);
		return context;
	}

	public void render() {
		var context = getOrInitializeContext();
		if (context == null) return;
		context.updateAndPaint();
	}

	public void keyPressed(InputKey keyCode, int scanCode, InputMods modifiers) {
		var context = getContext();
		if (context == null) return;
		context.keyPressed(keyCode, scanCode, modifiers);
	}

	public void charTyped(int codePoint) {
		var context = getContext();
		if (context == null) return;
		context.charTyped(codePoint);
	}

	public void keyReleased(InputKey keyCode, int scanCode, InputMods modifiers) {
		var context = getContext();
		if (context == null) return;
		context.keyReleased(keyCode, scanCode, modifiers);
	}

	public void mousePressed(double mouseX, double mouseY, int button) {
		var context = getContext();
		if (context == null) return;
		context.mousePressed(mouseX, mouseY, button);
	}

	public void mouseDragged(double mouseX, double mouseY, int button, double deltaX, double deltaY) {
		var context = getContext();
		if (context == null) return;
		context.mouseDragged(mouseX, mouseY, button, deltaX, deltaY);
	}

	public void mouseMoved(double mouseX, double mouseY) {
		var context = getContext();
		if (context == null) return;
		context.mouseMoved(mouseX, mouseY);
	}

	public void mouseReleased(double mouseX, double mouseY, int button) {
		var context = getContext();
		if (context == null) return;
		context.mouseReleased(mouseX, mouseY, button);
	}

	public void mouseScrolled(double mouseX, double mouseY, double delta) {
		var context = getContext();
		if (context == null) return;
		context.mouseScrolled(mouseX, mouseY, delta);
	}
}
